import ast
from functools import partial
from typing import overload, Union, List, Callable, Optional
from typing_extensions import Literal

from mr_proper.common_types import AnyFuncdef, PureCheckResult
from mr_proper.pure_validators import (
    has_no_blacklisted_calls, uses_only_args_and_local_vars,
    has_returns, not_mutates_args, not_has_local_imports, not_has_forbidden_arguments_types,
    not_uses_self_or_class_vars,
)
from mr_proper.utils.ast_pure import get_not_pure_internal_calls

PureValidatorType = Callable[[AnyFuncdef, Optional[ast.Module]], List[str]]


@overload
def is_function_pure(
    funcdef_node: AnyFuncdef,
    file_ast_tree: ast.Module,
    with_errors: Literal[False],
    recursive: bool = False,
    pyfilepath: str = None,
) -> bool:
    ...


@overload
def is_function_pure(
    funcdef_node: AnyFuncdef,
    file_ast_tree: ast.Module,
    with_errors: Literal[True],
    recursive: bool = False,
    pyfilepath: str = None,
) -> PureCheckResult:
    ...


def is_function_pure(
    funcdef_node: AnyFuncdef,
    file_ast_tree: ast.Module = None,
    with_errors: bool = False,
    recursive: bool = False,
    pyfilepath: str = None,
    extra_forbidden_argument_type_names: List[str] = None,
) -> Union[bool, PureCheckResult]:
    validators: List[PureValidatorType] = [
        has_no_blacklisted_calls,
        uses_only_args_and_local_vars,
        has_returns,
        not_mutates_args,
        not_has_local_imports,
        partial(
            not_has_forbidden_arguments_types,
            extra_forbidden_argument_type_names=extra_forbidden_argument_type_names,
        ),
        not_uses_self_or_class_vars,
    ]

    errors: List[str] = []
    is_pure = True
    for validator_callable in validators:
        validator_errors = validator_callable(funcdef_node, file_ast_tree)
        if validator_errors:
            is_pure = False
            errors += list(set(validator_errors))

    if recursive and file_ast_tree and pyfilepath:
        for dirty_call_name in get_not_pure_internal_calls(funcdef_node, file_ast_tree, pyfilepath):
            errors.append(f'it calls for non-pure function ({dirty_call_name})')

    return (is_pure, errors) if with_errors else is_pure



import ast
import inspect
import importlib.util
from typing import Optional, List, cast

from mypy_extensions import TypedDict

from mr_proper.common_types import AnyFuncdef
from mr_proper.utils.ast import BUILTINS_LIST, get_nodes_from_funcdef_body, get_ast_tree


class EntityImportInfo(TypedDict):
    file_path: Optional[str]
    import_path: Optional[str]
    name: str


def get_not_pure_internal_calls(
    funcdef_node: AnyFuncdef,
    file_ast_tree: ast.Module,
    pyfilepath: str,
) -> List[str]:
    from mr_proper.public_api import is_function_pure

    not_pure_calls: List[str] = []
    for call_node in get_nodes_from_funcdef_body(funcdef_node, [ast.Call]):
        if not isinstance(call_node.func, ast.Name):
            continue  # recursively check only functions calls, not methods/attrs calls
        if (
            call_node.func.id.lower() == call_node.func.id  # check for only snake_case calls
            and call_node.func.id not in BUILTINS_LIST
        ):
            import_info = get_name_import_path(call_node.func, pyfilepath)
            imported_funcdef_node = get_funcdef_by(import_info) if import_info else None
            if imported_funcdef_node and import_info:
                filepath = cast(str, import_info['file_path'])
                is_call_clean: Optional[bool] = is_function_pure(
                    imported_funcdef_node,
                    file_ast_tree=file_ast_tree,
                    recursive=False,
                    pyfilepath=filepath,
                    with_errors=False,
                )
            else:
                is_call_clean = None
            if is_call_clean is False:
                not_pure_calls.append(call_node.func.id)
    return not_pure_calls


def get_name_import_path(name_node: ast.Name, pyfilepath: str) -> Optional[EntityImportInfo]:
    current_node = name_node.parent  # type: ignore
    while True:
        for child in ast.iter_child_nodes(current_node):
            # check for Import, not only ImportFrom
            if isinstance(child, ast.ImportFrom) and name_node.id in (n.name for n in child.names):
                return extract_import_info_from_import_node(child, name_node)
            elif (
                isinstance(child, (ast.FunctionDef, ast.AsyncFunctionDef))
                and child.name == name_node.id
            ):
                return {
                    'file_path': pyfilepath if pyfilepath != 'built-in' else None,
                    'import_path': None,
                    'name': name_node.id,
                }
        current_node = current_node.parent
        if not current_node:
            break
    return None


def extract_import_info_from_import_node(
    import_from_node: ast.ImportFrom,
    name_node: ast.Name,
) -> EntityImportInfo:
    import_path = import_from_node.module
    if import_path is None:
        filepath = None
    else:
        filepath = get_file_path_by(import_path, name_node.id)
        if filepath == 'built-in':
            filepath = None
    return {'file_path': filepath, 'import_path': import_from_node.module, 'name': name_node.id}


def get_funcdef_by(import_info: EntityImportInfo) -> Optional[AnyFuncdef]:
    file_path = import_info['file_path']
    if file_path:
        ast_tree = get_ast_tree(file_path)
        if not ast_tree:
            return None
        try:
            funcdef = [
                n for n in ast.walk(ast_tree)
                if (
                    isinstance(n, (ast.FunctionDef, ast.AsyncFunctionDef))
                    and n.name == import_info['name']
                )
            ]
        except AttributeError:
            return None
        if funcdef:
            return funcdef[0]
    return None


def get_file_path_by(
    import_path: str,
    imported_name: str,
    dynamic_resolver: bool = False,
) -> Optional[str]:
    if dynamic_resolver:
        module = importlib.import_module(import_path)
        return inspect.getfile(getattr(module, imported_name))
    else:
        try:
            module_spec = importlib.util.find_spec(import_path)
        except ModuleNotFoundError:
            return None
        return module_spec.origin if module_spec else None



import os


CALLABLES_BLACKLIST = ['print', 'open', 'post']
ATTRIBUTES_BLACKLIST = [
    'objects', 'post', 'count', 'all', 'exists',
    'filter', 'values_list', 'values', 'save',
]
FORBIDDEN_ARGUMENT_TYPES = ['QuerySet', 'Model', 'Callable']
FORBIDDEN_ATTR_CALLS_FOR_ARGS = {
    # list mutating methods
    'append',
    'clear',
    'extend',
    'insert',
    'pop',
    'remove',
    'reverse',
    'sort',
    # dict mutation_methods
    'update',
}

TARGET_PYTHON_VERSION = os.environ.get('MR_PROPER_TARGET_PYTHON_VERSION', '3.7')



import sys
import argparse

from mr_proper.public_api import is_function_pure
from mr_proper.utils.ast import get_ast_tree, get_all_funcdefs_from


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument('filename', type=str)
    parser.add_argument('--recursive', action='store_true')
    return parser.parse_args()


def main() -> None:
    args = parse_args()
    ast_tree = get_ast_tree(args.filename)
    if not ast_tree:
        sys.stdout.write('Error parsing ast tree\n')
        return
    for funcdef_node in get_all_funcdefs_from(ast_tree):
        function_name = funcdef_node.name
        is_pure, pureness_errors = is_function_pure(
            funcdef_node,
            ast_tree,
            with_errors=True,
            recursive=args.recursive,
            pyfilepath=args.filename,
        )
        if is_pure:
            sys.stdout.write(f'{function_name} is pure!\n')
        else:
            sys.stdout.write(f'{function_name} is not pure because of:\n')
            for error in pureness_errors:
                sys.stdout.write(f'\t{error}\n')



import ast
from typing import List, Optional, Callable, cast

from mr_proper.common_types import AnyFuncdef
from mr_proper.config import (
    CALLABLES_BLACKLIST, ATTRIBUTES_BLACKLIST, FORBIDDEN_ARGUMENT_TYPES,
    FORBIDDEN_ATTR_CALLS_FOR_ARGS)
from mr_proper.utils.ast import (
    get_nodes_from_funcdef_body, is_imported_from_stdlib,
    get_local_var_names_from_funcdef,
    BUILTINS_LIST)
from mr_proper.utils.python_naming import is_python_class_name


def has_no_blacklisted_calls(
    funcdef_node: AnyFuncdef,
    file_ast_tree: Optional[ast.Module],
) -> List[str]:
    errors: List[str] = []
    for call_node in [n for n in ast.walk(funcdef_node) if isinstance(n, ast.Call)]:
        if isinstance(call_node.func, ast.Name) and call_node.func.id in CALLABLES_BLACKLIST:
            errors.append(f'it calls forbidden function ({call_node.func.id})')
    for attr_node in [n for n in ast.walk(funcdef_node) if isinstance(n, ast.Attribute)]:
        if attr_node.attr in ATTRIBUTES_BLACKLIST:
            errors.append(f'it accesses forbidden attribute ({attr_node.attr})')
    return errors


def uses_only_args_and_local_vars(
    funcdef_node: AnyFuncdef,
    file_ast_tree: Optional[ast.Module],
    allow_external_class_usage: bool = False,
    allow_stdlib_usage: bool = True,
) -> List[str]:
    all_names = [n.id for n in get_nodes_from_funcdef_body(funcdef_node, [ast.Name])]
    args_names = [a.arg for a in funcdef_node.args.args]
    called_names = [
        c.func.id
        for c in get_nodes_from_funcdef_body(funcdef_node, [ast.Call])
        if isinstance(c.func, ast.Name)
    ]
    local_vars_names = get_local_var_names_from_funcdef(funcdef_node)
    subfunctions_arguments = {
        a.arg
        for n in get_nodes_from_funcdef_body(funcdef_node, [ast.FunctionDef, ast.AsyncFunctionDef])
        for a in n.args.args  # type: ignore
    }

    nonlocal_names = list(
        set(all_names)
        - set(args_names)
        - set(local_vars_names)
        - set(called_names)
        - subfunctions_arguments
        - BUILTINS_LIST,
    )
    if allow_external_class_usage:
        nonlocal_names = [n for n in nonlocal_names if not is_python_class_name(n)]
    if allow_stdlib_usage and file_ast_tree:
        nonlocal_names = [
            n for n in nonlocal_names
            if not is_imported_from_stdlib(n, file_ast_tree)
        ]

    return [f'it uses external name ({n})' for n in nonlocal_names]


def has_returns(funcdef_node: AnyFuncdef, file_ast_tree: Optional[ast.Module]) -> List[str]:
    if not get_nodes_from_funcdef_body(funcdef_node, [ast.Return, ast.Yield]):
        return ['it has no returns']
    return []


def not_mutates_args(funcdef_node: AnyFuncdef, file_ast_tree: Optional[ast.Module]) -> List[str]:
    args_names = {a.arg for a in funcdef_node.args.args}
    mutating_actions = [  # noqa
        (  # n +=
            ast.AugAssign,
            lambda n: True,
            lambda n: n.target,
        ),
        (  # n[...] =
            ast.Assign,
            lambda n: (
                n.targets
                and isinstance(n.targets[0], ast.Subscript)
                and isinstance(n.targets[0].value, ast.Name)
                and n.targets[0].value.id in args_names
            ),
            lambda n: n.targets[0],
        ),
        (  # n.foo =
            ast.Assign,
            lambda n: n.targets and isinstance(n.targets[0], ast.Attribute),
            lambda n: n.targets[0].value,
        ),
        (  # n.append/remove/...(...)
            ast.Attribute,
            lambda n: (
                n.attr in FORBIDDEN_ATTR_CALLS_FOR_ARGS
                and isinstance(n.value, ast.Name)
                and n.value.id in args_names
            ),
            lambda n: n.value,
        ),
    ]

    errors: List[str] = []
    for node_type, node_validator, node_extractor in mutating_actions:
        node_validator = cast(Callable[[ast.AST], bool], node_validator)
        node_extractor = cast(Callable[[ast.AST], ast.AST], node_extractor)
        for node in get_nodes_from_funcdef_body(funcdef_node, [node_type]):
            if not node_validator(node):
                continue
            for name_node in [n for n in ast.walk(node_extractor(node)) if isinstance(n, ast.Name)]:
                if name_node.id in args_names:
                    errors.append(f'it mutates its own argument (on line {node.lineno})')
    return errors


def not_has_local_imports(
    funcdef_node: AnyFuncdef,
    file_ast_tree: Optional[ast.Module],
) -> List[str]:
    if get_nodes_from_funcdef_body(funcdef_node, [ast.ImportFrom, ast.Import]):
        return ['it has local imports']
    return []


def not_has_forbidden_arguments_types(
    funcdef_node: AnyFuncdef,
    file_ast_tree: Optional[ast.Module],
    extra_forbidden_argument_type_names: List[str] = None,
) -> List[str]:
    errors: List[str] = []
    forbidden_argument_type_names = FORBIDDEN_ARGUMENT_TYPES
    if extra_forbidden_argument_type_names:
        forbidden_argument_type_names += extra_forbidden_argument_type_names
    for argument in funcdef_node.args.args:
        for annotation_part_node in ast.walk(argument.annotation) if argument.annotation else []:
            if not isinstance(annotation_part_node, ast.Name):
                continue
            type_without_prefix = annotation_part_node.id.split('.')[-1]
            if type_without_prefix in forbidden_argument_type_names:
                errors.append(f'it has type {type_without_prefix} in argument types')
    return errors


def not_uses_self_or_class_vars(
    funcdef_node: AnyFuncdef,
    file_ast_tree: Optional[ast.Module],
) -> List[str]:
    forbidden_names = {'self', 'cls', 'super'}

    errors: List[str] = []
    for name_node in get_nodes_from_funcdef_body(funcdef_node, [ast.Name]):
        if name_node.id in forbidden_names:
            errors.append(f'it uses {name_node.id} var')
    return errors



import ast
import builtins
from typing import Optional, List, Union, Type, TypeVar, cast

from stdlib_list import stdlib_list

from mr_proper.common_types import AnyFuncdef
from mr_proper.config import TARGET_PYTHON_VERSION


T = TypeVar('T', bound=ast.AST)


BUILTINS_LIST = {b for b in dir(builtins) if not b.startswith('_')}
STDLIB_MODULES_NAMES = stdlib_list(TARGET_PYTHON_VERSION)


def get_ast_tree(pyfilepath: str) -> Optional[ast.Module]:
    with open(pyfilepath, 'r') as file_handler:
        try:
            file_content = file_handler.read()
        except UnicodeDecodeError:
            return None
    try:
        ast_tree = ast.parse(file_content)
    except SyntaxError:
        return None
    set_parents(ast_tree)
    return ast_tree


def set_parents(ast_tree: ast.AST) -> None:
    for node in ast.walk(ast_tree):
        for child in ast.iter_child_nodes(node):
            child.parent = node  # type: ignore
    ast_tree.parent = None  # type: ignore


def get_all_funcdefs_from(ast_tree: ast.AST) -> List[AnyFuncdef]:
    return [n for n in ast.walk(ast_tree) if isinstance(n, (ast.FunctionDef, ast.AsyncFunctionDef))]


def get_nodes_from_funcdef_body(
    funcdef: Union[ast.FunctionDef, ast.AsyncFunctionDef],
    of_types: List[Type[T]],
) -> List[T]:
    nodes: List[ast.AST] = []
    for node in funcdef.body:
        nodes += [n for n in ast.walk(node) if isinstance(n, tuple(of_types))]
    return nodes  # type: ignore


def is_imported_from_stdlib(name: str, file_ast_tree: ast.Module) -> Optional[bool]:
    for import_node in get_all_global_import_nodes(file_ast_tree):
        for full_import_path in get_full_import_pathes(import_node):
            if full_import_path.split('.')[-1] == name:
                base_module = full_import_path.split('.')[0]
                return base_module in STDLIB_MODULES_NAMES
    return None


def get_local_var_names_from_funcdef(funcdef_node: AnyFuncdef) -> List[str]:
    local_vars_names: List[str] = []
    for assign_node in get_nodes_from_funcdef_body(funcdef_node, [ast.Assign]):
        for target in assign_node.targets:
            local_vars_names += [n.id for n in ast.walk(target) if isinstance(n, ast.Name)]
    for annassign_node in get_nodes_from_funcdef_body(funcdef_node, [ast.AnnAssign]):
        if isinstance(annassign_node.target, ast.Name):
            local_vars_names.append(annassign_node.target.id)
    for comprehension in get_nodes_from_funcdef_body(funcdef_node, [ast.comprehension, ast.For]):
        comprehension = cast(ast.comprehension, comprehension)
        local_vars_names += get_local_var_names_from_loop(comprehension)
    local_vars_names += {
        n.name
        for n in ast.walk(funcdef_node)
        if isinstance(n, ast.ExceptHandler) and n.name
    }
    return list(set(local_vars_names))


def get_local_var_names_from_loop(loop_node: Union[ast.comprehension, ast.For]) -> List[str]:
    if isinstance(loop_node.target, ast.Name):
        return [loop_node.target.id]
    elif isinstance(loop_node.target, ast.Tuple):
        return [e.id for e in loop_node.target.elts if isinstance(e, ast.Name)]
    return []


def get_all_global_import_nodes(file_ast_tree: ast.AST) -> List[Union[ast.ImportFrom, ast.Import]]:
    nodes_with_local_scope = {ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef}

    import_nodes: List[Union[ast.ImportFrom, ast.Import]] = []
    for node in getattr(file_ast_tree, 'body', []):
        if isinstance(node, (ast.ImportFrom, ast.Import)):
            import_nodes.append(node)
        if hasattr(node, 'body') and not isinstance(node, tuple(nodes_with_local_scope)):
            import_nodes += get_all_global_import_nodes(node)
    return import_nodes


def get_full_import_pathes(import_node: Union[ast.ImportFrom, ast.Import]) -> List[str]:
    if isinstance(import_node, ast.Import):
        return [n.name for n in import_node.names]
    elif isinstance(import_node, ast.ImportFrom):
        return [f'{import_node.module}.{n.name}' for n in import_node.names]



from setuptools import setup, find_packages


package_name = 'mr_proper'


def get_long_description() -> str:
    with open('README.md', encoding='utf8') as f:
        return f.read()


setup(
    name=package_name,
    description='Package to find typos in russian text.',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    include_package_data=True,
    keywords='typos',
    version='0.0.3',
    author='Ilya Lebedev',
    author_email='melevir@gmail.com',
    install_requires=[
        'setuptools',
        'stdlib-list>=0.5.0',
        'mypy-extensions==0.4.1',
    ],
    entry_points={
        'console_scripts': [
            'mr_proper = mr_proper.main:main',
        ],
    },
    url='https://github.com/best-doctor/mr_proper',
    license='MIT',
    py_modules=[package_name],
    zip_safe=False,
)
